"""Provide utility functions for RPC."""

import ast
import asyncio
import contextlib
import copy
import inspect
import io
import re
import secrets
import string
import traceback
import collections.abc
from functools import partial
from inspect import Parameter, Signature
from types import BuiltinFunctionType, FunctionType
from typing import Any


def generate_password(length=50):
    """Generate a password."""
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for i in range(length))


_hash_id = generate_password()


def recursive_hash(obj):
    """Generate a hash for nested dictionaries and lists."""
    if isinstance(obj, collections.abc.Hashable) and not isinstance(obj, dotdict):
        return hash(obj)
    elif isinstance(obj, dict) or isinstance(obj, dotdict):
        return hash(tuple(sorted((k, recursive_hash(v)) for k, v in obj.items())))
    elif isinstance(obj, (list, tuple)):
        return hash(tuple(recursive_hash(i) for i in obj))
    else:
        raise TypeError(f"Unsupported type: {type(obj)}")


class dotdict(dict):  # pylint: disable=invalid-name
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.__getitem__
    __delattr__ = dict.__delitem__

    def __init__(self, *args, **kwargs):
        """Initialize the dotdict, recursively making all nested dicts dotdicts."""
        super().__init__(*args, **kwargs)

    def __setattr__(self, name, value):
        """Set the attribute."""
        if isinstance(value, dict) and not isinstance(value, dotdict):
            value = dotdict(value)
        super().__setitem__(name, value)

    def __hash__(self):
        """Return the hash."""
        return hash(tuple(sorted(self.items())))

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))

    def __getattribute__(self, name):
        if name in self:
            value = self[name]
            if isinstance(value, dict) and not isinstance(value, dotdict):
                value = dotdict(value)
            return value
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return None

    def update(self, other=None, **kwargs):
        """Update the dictionary with the key/value pairs from other, recursively merging nested dictionaries."""
        if other is not None:
            if isinstance(other, dict):
                for key, value in other.items():
                    if (
                        isinstance(value, dict)
                        and key in self
                        and isinstance(self[key], dict)
                    ):
                        self[key].update(value)
                    else:
                        if isinstance(value, dict) and not isinstance(value, dotdict):
                            value = dotdict(value)
                        self[key] = value
            else:
                raise TypeError(
                    "update() argument must be dict, not {}".format(
                        type(other).__name__
                    )
                )
        if kwargs:
            self.update(kwargs)


def format_traceback(traceback_string):
    """Format traceback."""
    formatted_lines = traceback_string.splitlines()
    # remove the second and third line
    formatted_lines.pop(1)
    formatted_lines.pop(1)
    formatted_error_string = "\n".join(formatted_lines)
    formatted_error_string = formatted_error_string.replace(
        'File "<string>"', "Plugin script"
    )
    return formatted_error_string


class MessageEmitter:
    """Represent a message emitter."""

    def __init__(self, logger=None):
        """Set up instance."""
        self._event_handlers = {}
        self._logger = logger

    def on(self, event, handler):
        """Register an event handler."""
        if event not in self._event_handlers:
            self._event_handlers[event] = []
        self._event_handlers[event].append(handler)

    def once(self, event, handler):
        """Register an event handler that should only run once."""

        # wrap the handler function,
        # this is needed because setting property
        # won't work for member function of a class instance
        def wrap_func(*args, **kwargs):
            return handler(*args, **kwargs)

        wrap_func.___event_run_once = True
        self.on(event, wrap_func)

    def off(self, event=None, handler=None):
        """Reset one or all event handlers."""
        if event is None and handler is None:
            self._event_handlers = {}
        elif event is not None and handler is None:
            if event in self._event_handlers:
                self._event_handlers[event] = []
        else:
            if event in self._event_handlers:
                self._event_handlers[event].remove(handler)

    def emit(self, msg):
        """Emit a message."""
        raise NotImplementedError

    def _fire(self, event, data=None):
        """Fire an event handler."""
        if event in self._event_handlers:
            for handler in self._event_handlers[event]:
                try:
                    ret = handler(data)
                    if inspect.isawaitable(ret):
                        asyncio.ensure_future(ret)
                except Exception as err:
                    traceback_error = traceback.format_exc()
                    if self._logger:
                        self._logger.exception(err)
                    self.emit({"type": "error", "message": traceback_error})
                finally:
                    if hasattr(handler, "___event_run_once"):
                        self._event_handlers[event].remove(handler)
        else:
            if self._logger and self._logger.debug:
                self._logger.debug("Unhandled event: {}, data: {}".format(event, data))


def encode_zarr_store(zobj):
    """Encode the zarr store."""
    import zarr

    path_prefix = f"{zobj.path}/" if zobj.path else ""

    def getItem(key, options=None):
        return zobj.store[path_prefix + key]

    def setItem(key, value):
        zobj.store[path_prefix + key] = value

    def containsItem(key, options=None):
        if path_prefix + key in zobj.store:
            return True

    return {
        "_rintf": True,
        "_rtype": "zarr-array" if isinstance(zobj, zarr.Array) else "zarr-group",
        "getItem": getItem,
        "setItem": setItem,
        "containsItem": containsItem,
    }


def register_default_codecs(api, options=None):
    """Register default codecs."""

    if options is None or "zarr-array" in options:
        import zarr

        api.registerCodec(
            {"name": "zarr-array", "type": zarr.Array, "encoder": encode_zarr_store}
        )

    if options is None or "zarr-group" in options:
        import zarr

        api.registerCodec(
            {"name": "zarr-group", "type": zarr.Group, "encoder": encode_zarr_store}
        )


def extract_function_info(func):
    """Extract function info."""
    # Create an in-memory text stream
    f = io.StringIO()

    # Redirect the output of help to the text stream
    with contextlib.redirect_stdout(f):
        help(func)
    help_string = f.getvalue()
    match = re.search(r"(\w+)\((.*?)\)\n\s*(.*)", help_string, re.DOTALL)
    if match:
        func_name, func_signature, docstring = match.groups()
        # Clean up the docstring
        docstring = func.__doc__ or re.sub(r"\n\s*", " ", docstring).strip()
        return {"name": func_name, "sig": func_signature, "doc": docstring}
    else:
        return None


def callable_sig(any_callable, skip_context=False):
    """Return the signature of a callable."""
    try:
        if isinstance(any_callable, partial):
            signature = inspect.signature(any_callable.func)
            name = any_callable.func.__name__
            fixed = set(any_callable.keywords)
        elif inspect.isclass(any_callable):
            signature = inspect.signature(any_callable.__call__)
            name = any_callable.__name__
            fixed = set()
        elif hasattr(any_callable, "__call__") and not isinstance(
            any_callable, (FunctionType, BuiltinFunctionType)
        ):
            signature = inspect.signature(any_callable)
            name = type(any_callable).__name__
            fixed = set()
        else:
            signature = inspect.signature(any_callable)
            name = any_callable.__name__
            fixed = set()
    except ValueError:
        # Provide a default signature for built-in functions
        signature = Signature(
            parameters=[
                Parameter(name="args", kind=Parameter.VAR_POSITIONAL),
                Parameter(name="kwargs", kind=Parameter.VAR_KEYWORD),
            ]
        )
        name = any_callable.__name__
        fixed = set()

    if skip_context:
        fixed.add("context")

    params = [p for name, p in signature.parameters.items() if name not in fixed]
    signature = Signature(parameters=params)

    # Remove invalid characters from name
    # e.g. <lambda> -> lambda
    name = re.sub(r"\W", "", name)

    primitive = True
    for p in signature.parameters.values():
        if (
            p.default is not None
            and p.default != inspect._empty
            and not isinstance(p.default, (str, int, float, bool, list, dict, tuple))
        ):
            primitive = False
    if primitive:
        sig_str = str(signature)
    else:
        sig_str = f"({', '.join([p.name for p in signature.parameters.values()])})"
    return f"{name}{sig_str}"


def callable_doc(any_callable):
    """Return the docstring of a callable."""
    if isinstance(any_callable, partial):
        return any_callable.func.__doc__

    try:
        return any_callable.__doc__
    except AttributeError:
        return None



import inspect
from typing import get_origin, get_args, Union
from functools import wraps, partial
from inspect import Signature, Parameter, signature
from typing import Any, Dict

try:
    from pydantic import create_model, Field, BaseModel
    from pydantic.fields import FieldInfo
    from pydantic_core import PydanticUndefined

    PYDANTIC_AVAILABLE = True
except ImportError:
    PYDANTIC_AVAILABLE = False


class Field:
    def __init__(self, default=inspect._empty, description=None):
        self.default = default
        self.description = description


# https://stackoverflow.com/a/58938747
def remove_a_key(d, remove_key):
    if isinstance(d, dict):
        for key in list(d.keys()):
            if key == remove_key:
                del d[key]
            else:
                remove_a_key(d[key], remove_key)


def get_type_name(tp):
    """Get the JSON schema type name for a given type."""
    if tp in [int, float, str, bool, list, dict]:
        return tp.__name__
    return {}  # Return an empty schema for non-primitive types


def extract_parameter_schema(param):
    """Extract the schema for a given parameter."""
    if param.annotation != inspect._empty:
        param_type = get_type_name(param.annotation)
    else:
        param_type = {}

    param_schema = {"type": param_type} if param_type else {}

    if param.default != inspect._empty:
        if isinstance(param.default, (int, float, str, bool, list, dict, type(None))):
            param_schema["default"] = param.default
        if isinstance(param.default, Field):
            param_schema["description"] = param.default.description

    return param_schema


def schema_function_native(
    original_func, name=None, schema_mode="strict", skip_context=False
):
    """Decorator to add input/output schema to a function."""
    assert callable(original_func)
    if hasattr(original_func, "__schema__"):
        return original_func

    if isinstance(original_func, partial):
        func_name = name or (original_func.func.__name__)
        func_doc = original_func.func.__doc__
        original_func_sig = signature(original_func.func)
        skip_args = set(original_func.keywords.keys())
    else:
        func_name = name or original_func.__name__
        func_doc = original_func.__doc__
        original_func_sig = signature(original_func)
        skip_args = set()

    if schema_mode == "strict":
        assert (
            func_name != "<lambda>"
        ), f"Lambda functions are not supported (name: {name})"

    parameters = [
        (name, param)
        for name, param in original_func_sig.parameters.items()
        if name not in skip_args
    ]

    if skip_context and parameters:
        parameters = parameters[:-1]  # Skip the last parameter if skip_context is True

    func_schema = {
        "type": "object",
        "properties": {
            name: extract_parameter_schema(param) for name, param in parameters
        },
        "required": [
            name for name, param in parameters if param.default == inspect._empty
        ],
    }

    if inspect.iscoroutinefunction(original_func):

        @wraps(original_func)
        async def wrapper(*args, **kwargs):
            return await original_func(*args, **kwargs)

    else:

        @wraps(original_func)
        def wrapper(*args, **kwargs):
            return original_func(*args, **kwargs)

    wrapper.__schema__ = {
        "name": func_name,
        "description": func_doc,
        "parameters": func_schema,
    }
    return wrapper


def dict_to_pydantic_model(
    name: str, dict_def: Dict[str, Any], doc: str = None
):
    fields = {}

    for field_name, value in dict_def.items():
        if isinstance(value, tuple):
            if len(value) == 2 and isinstance(value[1], Field):
                # Handle the case where value is a tuple (type, Field(...))
                field_type, field_info = value
                fields[field_name] = (
                    field_type,
                    field_info.default if field_info.default is not Ellipsis else ...,
                )
            else:
                fields[field_name] = value
        elif isinstance(value, dict):
            fields[field_name] = (
                dict_to_pydantic_model(f"{name}_{field_name}", value),
                ...,
            )
        else:
            raise ValueError(f"Field {field_name}:{value} has invalid syntax")

    model = create_model(name, **fields)
    model.__doc__ = doc
    return model


def extract_tool_schemas(
    func, func_name=None, schema_mode="strict", skip_context=False, skip_args=set()
):
    assert PYDANTIC_AVAILABLE, "Pydantic is not available"
    assert callable(func), "Tool function must be callable functions"
    sig = signature(func)
    func_name = func.__name__ if not isinstance(func, partial) else func.func.__name__

    names = [p.name for p in sig.parameters.values() if p.name not in skip_args]

    if skip_context and names:
        names = names[:-1]

    for name in names:
        assert (
            sig.parameters[name].annotation != inspect._empty or schema_mode == "auto"
        ), f"Argument `{name}` for `{func_name}` must have type annotation"

    types = [
        (
            sig.parameters[name].annotation
            if sig.parameters[name].annotation != inspect._empty
            else Any
        )
        for name in names
    ]
    defaults = []

    for name in names:
        if sig.parameters[name].default == inspect._empty:
            defaults.append(Field(..., description=name))
        else:
            if schema_mode == "strict":
                assert isinstance(
                    sig.parameters[name].default, FieldInfo
                ), "Argument default must be a FieldInfo object with description"
                assert (
                    sig.parameters[name].default.description is not None
                ), f"Argument `{name}` for `{func_name}` must have a description"
            default_value = (
                sig.parameters[name].default
                if isinstance(sig.parameters[name].default, FieldInfo)
                else Field(default=sig.parameters[name].default)
            )
            # check if the actual value is json serializable
            if default_value.default != PydanticUndefined and not isinstance(
                default_value.default, (int, float, str, bool, type(None))
            ):
                raise ValueError(
                    f"Argument `{name}` for `{func_name}` must have a default value that is json serializable"
                )
            defaults.append(default_value)

    func_name = func_name or func.__name__
    return (
        dict_to_pydantic_model(
            func_name,
            {names[i]: (types[i], defaults[i]) for i in range(len(names))},
            func.__doc__,
        ),
        sig.return_annotation,
    )


def schema_function_pydantic(
    original_func, input_model=None, name=None, schema_mode="strict", skip_context=False
):
    """Decorator to add input/output schema to a function."""
    assert PYDANTIC_AVAILABLE, "Pydantic is not available"
    assert callable(original_func)
    if hasattr(original_func, "__schema__"):
        return original_func

    if isinstance(original_func, partial):
        func_name = name or (original_func.func.__name__ + "(partial)")
        original_func_sig = signature(original_func.func)
        skip_args = set(original_func.keywords.keys())
    else:
        func_name = name or original_func.__name__
        original_func_sig = signature(original_func)
        skip_args = set()

    if schema_mode == "strict":
        assert func_name != "<lambda>", "Lambda functions are not supported"
        assert (
            original_func.__doc__ is not None
        ), f"Function `{func_name}` must have a docstring"

    if input_model:
        parameters = []
        for name, field in input_model.model_fields.items():
            if name not in skip_args:
                parameters.append(
                    Parameter(
                        name,
                        kind=Parameter.POSITIONAL_OR_KEYWORD,
                        annotation=field.annotation,
                        default=field,
                    )
                )
        func_sig = Signature(parameters)
    else:
        input_model, _ = extract_tool_schemas(
            original_func.func if isinstance(original_func, partial) else original_func,
            schema_mode=schema_mode,
            skip_context=skip_context,
            skip_args=skip_args,
        )
        func_sig = original_func_sig

    assert input_model is not None

    defaults = {}
    default_factories = {}
    required = []
    for name, field in input_model.model_fields.items():
        if field.default != PydanticUndefined or field.default_factory is not None:
            if field.default_factory is not None:
                default_factories[name] = field.default_factory
            else:
                defaults[name] = field.default
        else:
            if get_origin(field.annotation) is Union and type(None) in get_args(
                field.annotation
            ):
                defaults[name] = None
            else:
                required.append(name)

    def process_arguments(args, kwargs):
        # Create a map of parameter names to values from args
        arg_map = {name: value for name, value in zip(func_sig.parameters.keys(), args)}
        for req in required:
            assert (
                req in kwargs or req in arg_map
            ), f"Function call `{func_name}` missing required argument `{req}`"
        for k in default_factories:
            if k not in kwargs:
                kwargs[k] = default_factories[k]()
        for k in defaults:
            if k in kwargs and isinstance(kwargs[k], FieldInfo):
                kwargs[k] = defaults[k]

        # Convert dictionary inputs to Pydantic model if needed
        new_args = []
        for arg, param in zip(args, func_sig.parameters.values()):
            if (
                isinstance(param.default, FieldInfo)
                and isinstance(param.annotation, type)
                and issubclass(param.annotation, BaseModel)
            ):
                if isinstance(arg, dict):
                    arg = param.annotation(**arg)
            new_args.append(arg)

        new_kwargs = {}
        for k, v in kwargs.items():
            param = func_sig.parameters[k]
            if (
                isinstance(param.default, FieldInfo)
                and isinstance(param.annotation, type)
                and issubclass(param.annotation, BaseModel)
            ):
                if isinstance(v, dict):
                    v = param.annotation(**v)
            new_kwargs[k] = v
        return new_args, new_kwargs

    # check if original_func is a coroutine
    if inspect.iscoroutinefunction(original_func):

        @wraps(original_func)
        async def wrapper(*args, **kwargs):
            new_args, new_kwargs = process_arguments(args, kwargs)
            ret = await original_func(*new_args, **new_kwargs)
            return ret

    else:

        @wraps(original_func)
        def wrapper(*args, **kwargs):
            new_args, new_kwargs = process_arguments(args, kwargs)
            ret = original_func(*new_args, **new_kwargs)
            return ret

    spec = input_model.model_json_schema()
    remove_a_key(spec, "title")
    description = spec.get("description", original_func.__doc__)
    if "description" in spec:
        del spec["description"]
    wrapper.__schema__ = {
        "name": func_name,
        "description": description,
        "parameters": spec,
    }
    wrapper.__original__ = original_func
    return wrapper


def schema_function(func, name=None, schema_type="auto", skip_context=False):
    """Decorator to add input/output schema to a function."""
    assert schema_type in [
        "pydantic",
        "native",
        "auto",
    ], "Schema type must be 'pydantic' or 'native'"
    if schema_type == "auto":
        if PYDANTIC_AVAILABLE:
            schema_type = "pydantic"
            schema_mode = "auto"
        else:
            schema_type = "native"
    if schema_type == "pydantic":
        return schema_function_pydantic(
            func, name=name, schema_mode=schema_mode, skip_context=skip_context
        )
    elif schema_type == "native":
        return schema_function_native(
            func, name=name, schema_mode=schema_mode, skip_context=skip_context
        )
    else:
        raise ValueError(f"Invalid schema type: {schema_type}")


def parse_schema_function(
    obj, name=None, schema_type="native", schema_mode="strict", skip_context=False
):
    """Recursively convert callable with schema_function based on schema type."""
    assert schema_type in [
        "pydantic",
        "native",
        "auto",
    ], "Schema type must be 'pydantic' or 'native'"

    if schema_type == "auto":
        if PYDANTIC_AVAILABLE:
            schema_type = "pydantic"
            schema_mode = "auto"
        else:
            schema_type = "native"
    if callable(obj):
        try:
            if schema_type == "pydantic":
                return schema_function_pydantic(
                    obj, name=name, schema_mode=schema_mode, skip_context=skip_context
                )
            elif schema_type == "native":
                return schema_function_native(
                    obj, name=name, schema_mode=schema_mode, skip_context=skip_context
                )
            else:
                raise Exception(f"Invalid schema type: {schema_type}")
        except ValueError as e:
            print(f"Error parsing schema for {name}: {e}")
            return obj
        
    elif isinstance(obj, dict):
        return {
            k: parse_schema_function(
                v,
                name=k,
                schema_mode=schema_mode,
                schema_type=schema_type,
                skip_context=skip_context,
            )
            for k, v in obj.items()
        }
    elif isinstance(obj, list):
        return [
            parse_schema_function(
                x,
                schema_mode=schema_mode,
                schema_type=schema_type,
                skip_context=skip_context,
            )
            for x in obj
        ]
    else:
        return obj



"""This module provides an interface to interact with the Hypha server synchronously."""

import asyncio
import inspect
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from functools import wraps

from hypha_rpc.utils import dotdict
from hypha_rpc.webrtc_client import get_rtc_service as get_rtc_service_async
from hypha_rpc.webrtc_client import (
    register_rtc_service as register_rtc_service_async,
)
from hypha_rpc.websocket_client import (
    connect_to_server as connect_to_server_async,
)
from hypha_rpc.websocket_client import normalize_server_url


def get_async_methods(instance):
    """Get all the asynchronous methods from the given instance."""
    methods = []
    for attr_name in dir(instance):
        if not attr_name.startswith("_"):
            attr_value = getattr(instance, attr_name)
            if inspect.iscoroutinefunction(attr_value):
                methods.append(attr_name)
    return methods


def convert_sync_to_async(sync_func, loop, executor):
    """Convert a synchronous function to an asynchronous function."""
    if asyncio.iscoroutinefunction(sync_func):
        return sync_func

    @wraps(sync_func)
    async def wrapped_async(*args, **kwargs):
        result_future = loop.create_future()
        args = _encode_callables(args, convert_async_to_sync, loop, executor)
        kwargs = _encode_callables(kwargs, convert_async_to_sync, loop, executor)

        def run_and_set_result():
            try:
                result = sync_func(*args, **kwargs)
                loop.call_soon_threadsafe(result_future.set_result, result)
            except Exception as e:
                loop.call_soon_threadsafe(result_future.set_exception, e)

        executor.submit(run_and_set_result)
        result = await result_future
        obj = _encode_callables(result, convert_async_to_sync, loop, executor)
        return obj

    wrapped_async._sync = sync_func
    return wrapped_async


def convert_async_to_sync(async_func, loop, executor):
    """Convert an asynchronous function to a synchronous function."""

    @wraps(async_func)
    def wrapped_sync(*args, **kwargs):
        args = _encode_callables(args, convert_sync_to_async, loop, executor)
        kwargs = _encode_callables(kwargs, convert_sync_to_async, loop, executor)

        async def async_wrapper():
            obj = await async_func(*args, **kwargs)
            return _encode_callables(obj, convert_async_to_sync, loop, executor)

        result = asyncio.run_coroutine_threadsafe(async_wrapper(), loop).result()
        return result

    wrapped_sync._async = async_func
    return wrapped_sync


def _encode_callables(obj, wrap, loop, executor):
    """Encode callables in the given object to sync or async."""
    if isinstance(obj, dict):
        return dotdict(
            {k: _encode_callables(v, wrap, loop, executor) for k, v in obj.items()}
        )
    elif isinstance(obj, (list, tuple)):
        return [_encode_callables(item, wrap, loop, executor) for item in obj]
    elif callable(obj):
        return wrap(obj, loop, executor)
    else:
        return obj


class SyncHyphaServer:
    """A class to interact with the Hypha server synchronously."""

    def __init__(self, sync_max_workers=2):
        """Initialize the SyncHyphaServer."""
        self.loop = None
        self.thread = None
        self.server = None
        # Note: we need at least 2 workers to avoid deadlock
        self.executor = ThreadPoolExecutor(max_workers=sync_max_workers)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.server:
            self.server.disconnect()

    async def _connect(self, config):
        config["loop"] = self.loop
        self.server = await connect_to_server_async(config)
        skip_keys = ["register_codec"]
        skipped = {k: v for k, v in self.server.items() if k not in skip_keys}
        obj = _encode_callables(
            skipped, convert_async_to_sync, self.loop, self.executor
        )
        for k in skip_keys:
            obj[k] = self.server[k]
        for k, v in obj.items():
            setattr(self, k, v)

    def _start_loop(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        self.loop = asyncio.get_event_loop()
        self.loop.run_forever()


def connect_to_server(config):
    """Connect to the Hypha server synchronously."""
    server = SyncHyphaServer(sync_max_workers=config.get("sync_max_workers", 2))

    if not server.loop:
        server.thread = threading.Thread(target=server._start_loop, daemon=True)
        server.thread.start()

    while not server.loop or not server.loop.is_running():
        pass

    future = asyncio.run_coroutine_threadsafe(server._connect(config), server.loop)
    future.result()

    return server


def login(config):
    """Login to the Hypha server."""
    server_url = normalize_server_url(config.get("server_url"))
    service_id = config.get("login_service_id", "public/*:hypha-login")
    timeout = config.get("login_timeout", 60)
    callback = config.get("login_callback")

    server = connect_to_server(
        {"name": "initial login client", "server_url": server_url}
    )
    try:
        svc = server.get_service(service_id)
        assert svc, f"Service {service_id} not found on the server."
        context = svc.start()
        if callback:
            callback(context)
        else:
            print(f"Please open your browser and login at {context['login_url']}")

        return svc.check(context["key"], timeout)
    except Exception as error:
        raise error
    finally:
        server.disconnect()


def register_rtc_service(server, service_id, config=None):
    """Register an RTC service on the Hypha server."""
    assert isinstance(
        server, SyncHyphaServer
    ), "server must be an instance of SyncHyphaServer, please use hypha.sync.connect_to_server to create a server instance."
    future = asyncio.run_coroutine_threadsafe(
        register_rtc_service_async(
            server.server,
            service_id,
            _encode_callables(
                config, convert_sync_to_async, server.loop, server.executor
            ),
        ),
        server.loop,
    )
    future.result()


def get_rtc_service(server, service_id, config=None):
    """Get an RTC service from the Hypha server."""
    assert isinstance(
        server, SyncHyphaServer
    ), "server must be an instance of SyncHyphaServer, please use hypha.sync.connect_to_server to create a server instance."

    future = asyncio.run_coroutine_threadsafe(
        get_rtc_service_async(
            server.server,
            service_id,
            _encode_callables(
                config, convert_sync_to_async, server.loop, server.executor
            ),
        ),
        server.loop,
    )
    pc = future.result()
    for func in get_async_methods(pc):
        setattr(
            pc,
            func,
            convert_async_to_sync(getattr(pc, func), server.loop, server.executor),
        )
    return pc


if __name__ == "__main__":
    server_url = "https://hypha.aicell.io"
    server = connect_to_server({"server_url": server_url})

    services = server.list_services("public")
    print("Public services: #", len(services))

    def hello(name):
        """Say hello."""
        print("Hello " + name)
        print("Current thread id: ", threading.get_ident(), threading.current_thread())
        time.sleep(2)
        return "Hello " + name

    server.register_service(
        {
            "name": "Hello World",
            "id": "hello-world",
            "config": {
                "visibility": "protected",
                "run_in_executor": True,
            },
            "hello": hello,
        }
    )

    workspace = server.config.workspace
    token = server.generate_token()

    server2 = connect_to_server(
        {"server_url": server_url, "workspace": workspace, "token": token}
    )

    hello_svc = server2.get_service("hello-world")

    print("Calling hello world service...")
    print(hello_svc.hello("World"))

    server.disconnect()
    server2.disconnect()



import pytest
import copy
from hypha_rpc.utils import dotdict  # Adjust the import based on your file structure


def test_getattr():
    d = dotdict({"a": 1, "b": {"c": 2}})
    assert d.a == 1
    assert d.b.c == 2


def test_getattr_keyerror():
    d = dotdict({"a": 1})
    assert d.nonexistent is None


def test_setattr():
    d = dotdict()
    d.a = 1
    assert d["a"] == 1


def test_setattr_nested():
    d = dotdict()
    d.b = {"c": 2}
    assert isinstance(d.b, dotdict)
    assert d.b.c == 2


def test_delattr():
    d = dotdict({"a": 1, "b": 2})
    del d.a
    assert "a" not in d


def test_delattr_keyerror():
    d = dotdict({"a": 1})
    with pytest.raises(KeyError):
        del d.nonexistent


def test_deepcopy():
    d = dotdict({"a": 1, "b": {"c": 2}})
    d_copy = copy.deepcopy(d)
    assert d == d_copy
    assert d is not d_copy
    assert d.b == d_copy.b
    assert d.b is not d_copy.b


def test_nested_dotdict():
    d = dotdict({"a": {"b": {"c": 2}}})
    assert isinstance(d.a, dotdict)
    assert isinstance(d.a.b, dotdict)
    assert d.a.b.c == 2


def test_regular_dict_methods():
    d = dotdict({"a": 1, "b": 2})
    assert d.get("a") == 1
    assert d.get("nonexistent", "default") == "default"
    assert list(d.keys()) == ["a", "b"]
    assert list(d.values()) == [1, 2]
    assert list(d.items()) == [("a", 1), ("b", 2)]


def test_update():
    d = dotdict({"a": 1, "b": 2})
    d.update({"b": 3, "c": 4})
    assert d.b == 3
    assert d.c == 4


def test_update_nested():
    d = dotdict({"a": {"b": 1}})
    d.update({"a": {"c": 2}})
    assert d.a.b == 1
    assert d.a.c == 2


def test_key_collision():
    d = dotdict({"get": "custom_value", "update": "another_value", "keys": 99})
    assert d["get"] == "custom_value"
    assert d["update"] == "another_value"
    assert d["keys"] == 99
    assert d.keys == 99


def test_recursive_initialization():
    d = dotdict({"a": {"b": {"c": 2}}})
    assert isinstance(d.a, dotdict)
    assert isinstance(d.a.b, dotdict)
    assert d.a.b.c == 2


def test_recursive_setitem():
    d = dotdict()
    d["a"] = {"b": {"c": 2}}
    assert isinstance(d.a, dotdict)
    assert isinstance(d.a.b, dotdict)
    assert d.a.b.c == 2


if __name__ == "__main__":
    pytest.main()



"""Test the schema generation."""

import pytest
import json
from hypha_rpc import (
    connect_to_server,
)
from pydantic import BaseModel, Field
from python.hypha_rpc.utils.schema import schema_function

from . import WS_SERVER_URL


class UserInfo(BaseModel):
    """User information."""

    name: str = Field(..., description="name of the user")
    age: int = Field(..., description="age of the user")


@schema_function
def say_hello(
    info: UserInfo = Field(..., description="Information of the person to say hello to")
) -> str:
    """Say hello to a person."""
    return f"hello {info.name}, you are {info.age} years old"


def exchange_contact(
    info: UserInfo = Field(
        ..., description="Information of the person to say hello to"
    ),
    with_age: bool = Field(True, description="whether the age should be exposed"),
) -> dict:
    """Exchange contact information."""
    return UserInfo(name="Alice", age=30 if with_age else -1).model_dump()


@pytest.mark.asyncio
async def test_schema_function(websocket_server):
    """Test extract schema from functions."""
    exchange_contact_with_schema = schema_function(exchange_contact)

    expected = "hello John, you are 20 years old"
    assert say_hello({"name": "John", "age": 20}) == expected
    assert say_hello(UserInfo(name="John", age=20)) == expected

    assert (
        exchange_contact_with_schema({"name": "John", "age": 20})
        == UserInfo(name="Alice", age=30).model_dump()
    )

    # test workspace is an exception, so it can pass directly
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    await ws.register_service(
        {
            "name": "Hello World",
            "id": "hello-world",
            "description": "hello world service",
            "config": {
                "visibility": "protected",
                "run_in_executor": True,
            },
            "say_hello": say_hello,
            "exchange_contact": exchange_contact_with_schema,
        }
    )
    svc = await ws.get_service("hello-world")

    assert svc.say_hello.__schema__
    assert svc.exchange_contact.__schema__
    # check some fields for the json schema
    assert svc.say_hello.__schema__["name"] == "say_hello"
    assert svc.say_hello.__schema__ == {
        "name": "say_hello",
        "description": "Say hello to a person.",
        "parameters": {
            "$defs": {
                "UserInfo": {
                    "description": "User information.",
                    "properties": {
                        "name": {"description": "name of the user", "type": "string"},
                        "age": {"description": "age of the user", "type": "integer"},
                    },
                    "required": ["name", "age"],
                    "type": "object",
                }
            },
            "properties": {
                "info": {
                    "allOf": [{"$ref": "#/$defs/UserInfo"}],
                    "description": "Information of the person to say hello to",
                }
            },
            "required": ["info"],
            "type": "object",
        },
    }
    assert await svc.say_hello({"name": "John", "age": 20}) == expected
    assert (
        await svc.exchange_contact({"name": "John", "age": 20})
        == UserInfo(name="Alice", age=30).model_dump()
    )


@pytest.mark.asyncio
async def test_schema_service(websocket_server):
    """Test creating schema service."""
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    def exchange_contact_with_context(
        info: UserInfo = Field(
            ..., description="Information of the person to say hello to"
        ),
        with_age: bool = Field(True, description="whether the age should be exposed"),
        context=None,
    ) -> dict:
        """Exchange contact information."""
        return UserInfo(name="Alice", age=30 if with_age else -1).model_dump()

    svc = await ws.register_service(
        dict(
            id="test",
            name="Test Service",
            description="Services for testing.",
            config={
                "require_context": True,
                "visibility": "public",
                "schema_type": "pydantic",
            },
            exchange_contact=exchange_contact_with_context,
            inner_service={
                "exchange_contact": exchange_contact_with_context,
            },
        )
    )
    assert svc.exchange_contact["type"] == "function"
    assert svc.exchange_contact["function"] == {
        "name": "exchange_contact",
        "description": "Exchange contact information.",
        "parameters": {
            "$defs": {
                "UserInfo": {
                    "description": "User information.",
                    "properties": {
                        "name": {
                            "description": "name of the user",
                            "type": "string",
                        },
                        "age": {
                            "description": "age of the user",
                            "type": "integer",
                        },
                    },
                    "required": ["name", "age"],
                    "type": "object",
                }
            },
            "properties": {
                "info": {
                    "allOf": [{"$ref": "#/$defs/UserInfo"}],
                    "description": "Information of the person to say hello to",
                },
                "with_age": {
                    "default": True,
                    "description": "whether the age should be exposed",
                    "type": "boolean",
                },
            },
            "required": ["info"],
            "type": "object",
        },
    }
    test_service = await ws.get_service(svc.id)

    assert test_service.exchange_contact.__schema__


@pytest.mark.asyncio
async def test_schema_service_modes(websocket_server):
    """Test creating schema service with different modes."""
    ws = await connect_to_server({"name": "my app", "server_url": WS_SERVER_URL})

    # Pydantic mode
    svc_pydantic = await ws.register_service(
        dict(
            id="test_pydantic",
            name="Test Service Pydantic",
            description="Services for testing pydantic mode.",
            config={
                "require_context": True,
                "visibility": "public",
                "schema_type": "pydantic",
            },
            exchange_contact=exchange_contact,
            inner_service={
                "exchange_contact": exchange_contact,
            },
        )
    )
    assert svc_pydantic.exchange_contact["type"] == "function"
    assert svc_pydantic.exchange_contact["function"]["name"] == "exchange_contact"
    assert "parameters" in svc_pydantic.exchange_contact["function"]

    # native mode
    svc_native = await ws.register_service(
        dict(
            id="test_native",
            name="Test Service native",
            description="Services for testing native mode.",
            config={
                "require_context": True,
                "visibility": "public",
                "schema_type": "native",
            },
            exchange_contact=exchange_contact,
            inner_service={
                "exchange_contact": exchange_contact,
            },
        )
    )
    assert svc_native.exchange_contact["type"] == "function"
    assert svc_native.exchange_contact["function"]["name"] == "exchange_contact"
    assert "parameters" in svc_native.exchange_contact["function"]

    test_service_pydantic = await ws.get_service(svc_pydantic.id)
    assert test_service_pydantic.exchange_contact.__schema__

    test_service_native = await ws.get_service(svc_native.id)
    assert test_service_native.exchange_contact.__schema__ == {
        "name": "exchange_contact",
        "description": "Exchange contact information.",
        "parameters": {"type": "object", "properties": {"info": {}}, "required": []},
    }
    assert "name" in test_service_native.exchange_contact.__schema__
    assert "description" in test_service_native.exchange_contact.__schema__
    assert "parameters" in test_service_native.exchange_contact.__schema__
